#!/usr/bin/perl

use CGI::Carp qw(fatalsToBrowser set_message);

BEGIN
{
	sub handle_errors
	{
		my $error=shift;
		$error.="Saving your edits failed.\nMake sure to make a local copy of your text.";
		$error=~s/\n/\\n/g;
		$error=~s/&quot;/"/g; #"
		$error=~s/&lt;/</g;
		$error=~s/&gt;/>/g;

		print "<html><body><script type=\"text/javascript\">alert('$error');</script></body></html>";
	}
	set_message(\&handle_errors);
}

use strict;

use CGI;
use File::Copy;
use Fcntl ':flock';

my @locked_tiddlers=("MainMenu","SiteTitle","SiteSubtitle"); # list of uneditable tiddlers
my $max_edit_time=3600; # maximum time between edits for backup overwrites

my $query=new CGI;
my $action=$query->param("action");

die "Saving does not work with referer blocking" unless($ENV{HTTP_REFERER});

my ($file)=$ENV{HTTP_REFERER}=~/([^\/]+)(?:#|\?|$)/;
$file="index.html" unless($file);

die "Suspicious filename" if($file=~/[^a-zA-Z0-9_\-\. ]/);
die "Can only edit HTML files" unless($file=~/\.html?$/i);

if($action eq "save")
{
	my $title=$query->param("title");
	my $modifier=$query->param("modifier");
	my $modified=$query->param("modified");
	my $lastmod=$query->param("lastmod");
	my $body=$query->param("body");

	update_page($file,$title,$modifier,$modified,$lastmod,$body);
}
elsif($action eq "delete")
{
	my $title=$query->param("title");
	my $modifier=$query->param("modifier");
	my $lastmod=$query->param("lastmod");

	update_page($file,$title,$modifier,"",$lastmod,"");
}
else
{
	die "No action specified";
}

print "Content-Type: text/plain\n\n";



sub update_page($$$$$$)
{
	my ($file,$title,$modifier,$modified,$lastmod,$body)=@_;

	# check if this entry is locked
	die "This tiddler is locked" if(grep { $_ eq $title } @locked_tiddlers);

	# open and lock file
	open FILE,"+<$file" or die "Couldn't read file \"$file\"";
	flock FILE,LOCK_EX;

	my $contents; while(<FILE>) { $contents.=$_ } # read entire file
#	my $timestamp=make_date_string(time());

	# clean inputs
	$title=clean_string($title);
	$modifier=clean_string($modifier);
	$body=clean_string($body);

	# fix newlines
	$body=~s/\r\n/\n/g;

	# remove old entry
	$contents=~s:\<div([^\>]+?)id="store\Q$title\E"([^\>]*)\>.*?\</div\>\s*::si;
	my $attrs=$1.$2;
	my ($old_modified)=$attrs=~/modified\s*=\s*["']([0-9]+)["']/;

	# make sure someone else hasn't edited this entry
	die "This tiddler has been modified while you were editing it" if($lastmod and $old_modified and $lastmod!=$old_modified);

	if($body) # add the new entry, if it exists
	{
		$contents=~s:(\<div[^\>]+?id="storeArea"[^\>]*\>(\s*\<!--.*?--\>\s*)*):$1<div id="store$title" modified="$modified" modifier="$modifier">$body</div>\n\n:si;
	}

	# save file
	seek FILE,0,0;
	truncate FILE,0;
	print FILE $contents;

	# all work done, close file to release lock
	close FILE;

	# make a backup
	make_backup($file);
}

sub make_backup()
{
	my ($file)=@_;

	my ($base,$ext)=$file=~/^(.+)\.(html?)$/i;
	my $time=time();
	my $ip=$ENV{REMOTE_ADDR};

	my @pages=sort glob("$base.*.$ext*");
	my $latest=pop @pages;

	my ($latest_time,$latest_ip)=$latest=~/$base\.([0-9]+)\.([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)\.$ext/;

	unlink $latest if($latest_ip eq $ip and $time-$latest_time<$max_edit_time);

	my $archname="$base.$time.$ip.$ext";

	copy $file,$archname;

	my @htmlpages=glob("$base.*.$ext");

	foreach my $page (@htmlpages)
	{
		`gzip $page`;
	}
}

sub clean_string($)
{
	my ($str)=@_;

	$str=~s/^\s*//; # remove preceeding whitespace
	$str=~s/\s*$//; # remove traling whitespace

	$str=~s/&/&amp;/g;
	$str=~s/\</&lt;/g;
	$str=~s/\>/&gt;/g;
	$str=~s/"/&quot;/g; #"
	$str=~s/'/&#039;/g;
	$str=~s/,/&#44;/g;

	# repair unicode entities
	$str=~s/&amp;(\#[0-9]+;)/&$1/g;

	return $str;
}
